module ParseSTXml
  ( xmlParse
  ) where

import Prelude hiding (either,maybe,sequence)
import Maybe
import STypes
import LexXml
import ParseSTLib

xmlParse :: String -> Document
xmlParse = sanitycheck . papply document emptyST . xmlLex

sanitycheck :: Show p => [(a,s,[(p,t)])] -> a
sanitycheck [] = error "***Error at line 0: document not XML?"
sanitycheck ((x,_,[]):_) = x
sanitycheck ((x,_,s@((n,_):_)):xs) =
  error ("***Error at "++show n++": data beyond end of parsed document")


---- Symbol table stuff ----

type SymTab = [(String,String)]
emptyST :: SymTab
emptyST  = []

fst3 (a,_,_) = a
snd3 (_,a,_) = a
thd3 (_,_,a) = a


{--- Document simplification ----

class DePERef a where
    de_PERef :: a -> SymTab -> a
instance DePERef Document where
    de_PERef (Document p e) st = Document (de_PERef p) e
instance DePERef Prolog where
    de_PERef (Prolog x dtd) st = Prolog x (de_PERef dtd)
instance DePERef a => DePERef (Maybe a) where
    de_PERef Nothing st  = Nothing
    de_PERef (Just x) st = Just (de_PERef x)
instance DePERef a => DePERef [a] where
    de_PERef [] st     = []
    de_PERef (x:xs) st = de_PERef x : de_PERef xs
instance DePERef DocTypeDecl where
    de_PERef (DTD n eid ms) st = DTD n eid (de_PERef ms)
instance DePERef MarkupDecl where
    de_PERef (MarkupPE p _) st = Markup p (find p st >>= parse >>= return)
-}


---- Auxiliary Parsing Functions ----

#define PARSEARG(tok)  P (\st inp -> case inp of { \
                                      ((p,tok n):ts) -> [(n,st,ts)]; \
                                      ts -> [] } )
name :: Parser SymTab Token Name
name =  PARSEARG(TokName)

string, freetext :: Parser SymTab Token String
string =   PARSEARG(TokName)
freetext = PARSEARG(TokFreeText)

maybe :: Parser SymTab Token a -> Parser SymTab Token (Maybe a)
maybe p =
    ( p >>= return . Just) +++
    ( return Nothing)

either :: Parser SymTab Token a -> Parser SymTab Token b -> Parser SymTab Token (Either a b)
either p q =
    ( p >>= return . Left) +++
    ( q >>= return . Right)

word :: String -> Parser SymTab Token ()
word s = P (\st inp-> case inp of {
                       ((p,TokName n):ts) -> if s==n then [((),st,ts)]
                                             else [];
                       ((p,TokFreeText n):ts) -> if s==n then [((),st,ts)]
                                                 else [];
                       ts -> [] } )

posn :: Parser SymTab Token Posn
posn = P (\st inp-> case inp of {
                     ((p,_):_) -> [(p,st,inp)];
                     [] -> []; } )

nmtoken :: Parser SymTab Token NmToken
nmtoken = (string +++ freetext)

#define PEREF(cons,fn) \
   ( do {n <- pereference;                        \
         tr <- stquery (lookup n);                \
         case tr of {                             \
           Nothing -> return (cons n (error ("undefined PEReference %"++n))); \
           (Just tr) -> do {reparse (xmlLex tr);  \
                            v <- fn;              \
                            return (cons n v)}}} )


---- XML Parsing Functions ----

document :: Parser SymTab Token Document
document = do
    p <- prolog `elserror` "unrecognisable XML prolog"
    e <- element `elserror` "no toplevel document element"
    ms <- many misc
    return (Document p e)

comment :: Parser SymTab Token Comment
comment = do
    bracket (tok TokCommentOpen) freetext (tok TokCommentClose)

processinginstruction :: Parser SymTab Token ProcessingInstruction
processinginstruction = do
    tok TokPIOpen
    n <- string  `elserror` "processing instruction has no target"
    f <- freetext
    tok TokPIClose `elserror` "missing ?>"
    return (n, f)

cdsect :: Parser SymTab Token CDSect
cdsect = do
    tok TokSectionOpen
    bracket (tok (TokSection CDATAx)) chardata (tok TokSectionClose)

prolog :: Parser SymTab Token Prolog
prolog = do
    x <- maybe xmldecl
    many misc
    dtd <- maybe doctypedecl
    many misc
    return (Prolog x dtd)

xmldecl :: Parser SymTab Token XMLDecl
xmldecl = do
    tok TokPIOpen
    (word "xml" +++ word "XML")
    p <- posn
    s <- freetext
    tok TokPIClose `elserror` "missing ?> in <?xml ...?>"
    return ((fst3. head . papply aux emptyST . xmlReLex p) s)
  where
    aux = do
      v <- versioninfo  `elserror` "missing XML version info"
      e <- maybe encodingdecl
      s <- maybe sddecl
      return (XMLDecl v e s)

versioninfo :: Parser SymTab Token VersionInfo
versioninfo = do
    (word "version" +++ word "VERSION")
    tok TokEqual
    bracket (tok TokQuote) freetext (tok TokQuote)

misc :: Parser SymTab Token Misc
misc = 
    ( comment >>= return . Comment) +++
    ( processinginstruction >>= return . PI)

doctypedecl :: Parser SymTab Token DocTypeDecl
doctypedecl = do
    tok TokSpecialOpen
    tok (TokSpecial DOCTYPEx)
    n <- name
    eid <- maybe externalid
    es <- maybe (bracket (tok TokSqOpen)
                         (many markupdecl)
                         (tok TokSqClose))
    tok TokAnyClose  `elserror` "missing > in DOCTYPE decl"
    return (DTD n eid (case es of { Nothing -> []; Just e -> e }))

markupdecl :: Parser SymTab Token MarkupDecl
markupdecl =
    ( elementdecl >>= return . Element) +++
    ( attlistdecl >>= return . AttList) +++
    ( entitydecl >>= return . Entity) +++
    ( notationdecl >>= return . Notation) +++
    ( misc >>= return . MarkupMisc) +++
    PEREF(MarkupPE,markupdecl)

extsubset :: Parser SymTab Token ExtSubset
extsubset = do
    td <- maybe textdecl
    ds <- many extsubsetdecl
    return (ExtSubset td ds)

extsubsetdecl :: Parser SymTab Token ExtSubsetDecl
extsubsetdecl =
    ( markupdecl >>= return . ExtMarkupDecl) +++
    ( conditionalsect >>= return . ExtConditionalSect) +++
    PEREF(ExtPEReference,extsubsetdecl)

sddecl :: Parser SymTab Token SDDecl
sddecl = do
    (word "standalone" +++ word "STANDALONE")
    tok TokEqual `elserror` "missing = in 'standalone' decl"
    bracket (tok TokQuote)
            ( (word "yes" >> return True) +++
              (word "no" >> return False) `elserror`
              "'standalone' decl requires 'yes' or 'no' value" )
            (tok TokQuote)

element :: Parser SymTab Token Element
element = do
    tok TokAnyOpen
    (ElemTag n as) <- elemtag
    (( do tok TokEndClose
          return (Elem n as [])) +++
     ( do tok TokAnyClose
          cs <- many content
          m <- bracket (tok TokEndOpen) name (tok TokAnyClose)
          checkmatch n m
          return (Elem n as cs))
     `elserror` "missing > or /> in element tag")

checkmatch :: Name -> Name -> Parser SymTab Token ()
checkmatch n m =
  if n == m then return ()
  else do p <- posn
          error ("Error at "++show p++": tag <"++n++"> terminated by </"++m++">")

elemtag :: Parser SymTab Token ElemTag
elemtag = do
    n <- name `elserror` "malformed element tag"
    as <- many attribute
    return (ElemTag n as)

attribute :: Parser SymTab Token Attribute
attribute = do
    n <- name
    tok TokEqual `elserror` "missing = in attribute"
    v <- attvalue `elserror` "missing attvalue"
    return (n,v)

content :: Parser SymTab Token Content
content =
    ( element >>= return . CElem) +++
    ( chardata >>= return . CString) +++
    ( reference >>= return . CRef) +++
    ( cdsect >>= return . CString) +++
    ( misc >>= return . CMisc)

elementdecl :: Parser SymTab Token ElementDecl
elementdecl = do
    tok TokSpecialOpen
    tok (TokSpecial ELEMENTx)
    n <- name `elserror` "missing identifier in ELEMENT decl"
    c <- contentspec `elserror` "missing content spec in ELEMENT decl"
    tok TokAnyClose `elserror` "expected > terminating ELEMENT decl"
    return (ElementDecl n c)

contentspec :: Parser SymTab Token ContentSpec
contentspec =
    ( word "EMPTY" >> return EMPTY) +++
    ( word "ANY" >> return ANY) +++
    ( mixed >>= return . Mixed) +++
    ( cp >>= return . ContentSpec) +++
    PEREF(ContentPE,contentspec)

choice :: Parser SymTab Token [CP]
choice = do
    bracket (tok TokBraOpen)
            (cp `sepby1` (tok TokPipe))
            (tok TokBraClose)

sequence :: Parser SymTab Token [CP]
sequence = do
    bracket (tok TokBraOpen)
            (cp `sepby1` (tok TokComma))
            (tok TokBraClose)

cp :: Parser SymTab Token CP
cp =
    ( do n <- name
         m <- modifier
         return (TagName n m)) +++
    ( do ss <- sequence
         m <- modifier
         return (Seq ss m)) +++
    ( do cs <- choice
         m <- modifier
         return (Choice cs m)) +++
    PEREF(CPPE,cp)

modifier :: Parser SymTab Token Modifier
modifier =
    ( tok TokStar >> return Star) +++
    ( tok TokQuery >> return Query) +++
    ( tok TokPlus >> return Plus) +++
    ( return None)

mixed :: Parser SymTab Token Mixed
mixed = do
    tok TokBraOpen
    tok TokHash
    word "PCDATA"
    cont
  where
    cont = ( tok TokBraClose >> return PCDATA) +++
           ( do cs <- many ( do tok TokPipe
                                n <- name
                                return n)
                tok TokBraClose
                tok TokStar
                return (PCDATAplus cs))

attlistdecl :: Parser SymTab Token AttListDecl
attlistdecl = do
    tok TokSpecialOpen
    tok (TokSpecial ATTLISTx)
    n <- name `elserror` "missing identifier in ATTLIST"
    ds <- many attdef
    tok TokAnyClose `elserror` "missing > terminating ATTLIST"
    return (AttListDecl n ds)

attdef :: Parser SymTab Token AttDef
attdef = do
    n <- name
    t <- atttype `elserror` "missing attribute type in attlist defn"
    d <- defaultdecl
    return (AttDef n t d)

atttype :: Parser SymTab Token AttType
atttype =
    ( word "CDATA" >> return StringType) +++
    ( tokenizedtype >>= return . TokenizedType) +++
    ( enumeratedtype >>= return . EnumeratedType)

tokenizedtype :: Parser SymTab Token TokenizedType
tokenizedtype =
    ( word "ID" >> return ID) +++
    ( word "IDREF" >> return IDREF) +++
    ( word "IDREFS" >> return IDREFS) +++
    ( word "ENTITY" >> return ENTITY) +++
    ( word "ENTITIES" >> return ENTITIES) +++
    ( word "NMTOKEN" >> return NMTOKEN) +++
    ( word "NMTOKENS" >> return NMTOKENS)

enumeratedtype :: Parser SymTab Token EnumeratedType
enumeratedtype =
    ( notationtype >>= return . NotationType) +++
    ( enumeration >>= return . Enumeration)

notationtype :: Parser SymTab Token NotationType
notationtype = do
    word "NOTATION"
    bracket (tok TokBraOpen)
            (name `sepby1` (tok TokPipe))
            (tok TokBraClose)

enumeration :: Parser SymTab Token Enumeration
enumeration =
    bracket (tok TokBraOpen)
            (nmtoken `sepby1` (tok TokPipe))
            (tok TokBraClose)

defaultdecl :: Parser SymTab Token DefaultDecl
defaultdecl =
    ( tok TokHash >> word "REQUIRED" >> return REQUIRED) +++
    ( tok TokHash >> word "IMPLIED" >> return IMPLIED) +++
    ( do f <- maybe (tok TokHash >> word "FIXED" >> return FIXED)
         a <- attvalue
         return (Default a f))

conditionalsect :: Parser SymTab Token ConditionalSect
conditionalsect =
    ( do tok TokSectionOpen
         tok (TokSection INCLUDEx)
         tok TokSqOpen `elserror` "missing [ after INCLUDE"
         i <- extsubsetdecl `elserror` "missing ExtSubsetDecl in INCLUDE"
         tok TokSectionClose `elserror` "missing ] after INCLUDE"
         return (IncludeSect i)) +++
    ( do tok TokSectionOpen
         tok (TokSection IGNOREx)
         tok TokSqOpen `elserror` "missing [ after IGNORE"
         i <- many ignoresectcontents
         tok TokSectionClose `elserror` "missing ] after IGNORE"
         return (IgnoreSect i))

ignoresectcontents :: Parser SymTab Token IgnoreSectContents
ignoresectcontents = do
    i <- ignore
    is <- many (do tok TokSectionOpen
                   ic <- ignoresectcontents
                   tok TokSectionClose
                   ig <- ignore
                   return (ic,ig))
    return (IgnoreSectContents i is)

ignore :: Parser SymTab Token Ignore
ignore = freetext >>= return . Ignore

reference :: Parser SymTab Token Reference
reference =
    ( entityref >>= return . RefEntity) +++
    ( charref >>= return . RefChar)

entityref :: Parser SymTab Token EntityRef
entityref = do
    n <- bracket (tok TokAmp) name (tok TokSemi)
    return (EntityRef n)

charref :: Parser SymTab Token CharRef
charref = do
    n <- bracket (tok TokAmp) freetext (tok TokSemi)
    return (CharRef n)

pereference :: Parser SymTab Token PEReference
pereference = do
    bracket (tok TokPercent) nmtoken (tok TokSemi)

entitydecl :: Parser SymTab Token EntityDecl
entitydecl =
    ( gedecl >>= return . EntityGEDecl) +++
    ( pedecl >>= return . EntityPEDecl)

gedecl :: Parser SymTab Token GEDecl
gedecl = do
    tok TokSpecialOpen
    tok (TokSpecial ENTITYx)
    n <- name
    e <- entitydef `elserror` "missing entity defn in G ENTITY decl"
    tok TokAnyClose `elserror` "expected > terminating G ENTITY decl"
    return (GEDecl n e)

pedecl :: Parser SymTab Token PEDecl
pedecl = do
    tok TokSpecialOpen
    tok (TokSpecial ENTITYx)
    tok TokPercent
    n <- name
    e <- pedef `elserror` "missing entity defn in P ENTITY decl"
    tok TokAnyClose `elserror` "expected > terminating P ENTITY decl"
    return (PEDecl n e)

entitydef :: Parser SymTab Token EntityDef
entitydef =
    ( entityvalue >>= return . DefEntityValue) +++
    ( do eid <- externalid
         ndd <- maybe ndatadecl
         return (DefExternalID eid ndd))

pedef :: Parser SymTab Token PEDef
pedef =
    ( entityvalue >>= return . PEDefEntityValue) +++
    ( externalid >>= return . PEDefExternalID)

externalid :: Parser SymTab Token ExternalID
externalid =
    ( do word "SYSTEM"
         s <- systemliteral
         return (SYSTEM s)) +++
    ( do word "PUBLIC"
         p <- pubidliteral
         s <- systemliteral
         return (PUBLIC p s))

ndatadecl :: Parser SymTab Token NDataDecl
ndatadecl = do
    word "NDATA"
    n <- name
    return (NDATA n)

textdecl :: Parser SymTab Token TextDecl
textdecl = do
    tok TokPIOpen
    (word "xml" +++ word "XML")
    v <- maybe versioninfo
    e <- encodingdecl
    tok TokPIClose `elserror` "expected ?> terminating text decl"
    return (TextDecl v e)

extparsedent :: Parser SymTab Token ExtParsedEnt
extparsedent = do
    t <- maybe textdecl
    c <- content
    return (ExtParsedEnt t c)

extpe :: Parser SymTab Token ExtPE
extpe = do
    t <- maybe textdecl
    e <- extsubsetdecl
    return (ExtPE t e)

encodingdecl :: Parser SymTab Token EncodingDecl
encodingdecl = do
    (word "encoding" +++ word "ENCODING")
    tok TokEqual `elserror` "expected = in 'encoding' decl"
    f <- bracket (tok TokQuote) freetext (tok TokQuote)
    return (EncodingDecl f)

notationdecl :: Parser SymTab Token NotationDecl
notationdecl = do
    tok TokSpecialOpen
    word "NOTATION"
    n <- name
    e <- either externalid publicid
    tok TokAnyClose `elserror` "expected > terminating NOTATION decl"
    return (NOTATION n e)

publicid :: Parser SymTab Token PublicID
publicid = do
    word "PUBLICID"
    p <- pubidliteral
    return (PUBLICID p)

entityvalue :: Parser SymTab Token EntityValue
entityvalue = do
    evs <- bracket (tok TokQuote) (many ev) (tok TokQuote)
    return (EntityValue evs)

ev :: Parser SymTab Token EV
ev =
    ( freetext >>= return . EVString) +++
    PEREF(EVPERef,ev) +++
    ( reference >>= return . EVRef)

attvalue :: Parser SymTab Token AttValue
attvalue = do
    avs <- bracket (tok TokQuote)
                   (many (either freetext reference))
                   (tok TokQuote)
    return (AttValue avs)

systemliteral :: Parser SymTab Token SystemLiteral
systemliteral = do
    s <- bracket (tok TokQuote) freetext (tok TokQuote)
    return (SystemLiteral s)		-- note: need to fold &...; escapes

pubidliteral :: Parser SymTab Token PubidLiteral
pubidliteral = do
    s <- bracket (tok TokQuote) freetext (tok TokQuote)
    return (PubidLiteral s)		-- note: need to fold &...; escapes

chardata :: Parser SymTab Token CharData
chardata = freetext -- >>= return . CharData

